from utils import *
from gui import *

from . import Car

class CarEngine:
    @staticmethod
    def getSmoothShiftProgress(shiftProgress: float):
        return math.sin(shiftProgress / 2 * math.pi)
    
    @staticmethod
    def getGearRatioTransition(previousGear: int, gear: int, shiftProgress: float, gearRatios: list[float]):
        smoothShiftProgress = CarEngine.getSmoothShiftProgress(shiftProgress)
        gearRatioTransition = map(smoothShiftProgress, 0, 1, gearRatios[previousGear], gearRatios[gear])
        return gearRatioTransition
    
    @staticmethod
    def getSpeedFromRpm(engineRpm: float, gearRatio: float, finalDrive: float, tireDiameter: float):
        wheelRpm = engineRpm / gearRatio / finalDrive
        wheelRps = wheelRpm / 60
        wheelSpeed = wheelRps * (math.pi * tireDiameter)
        return wheelSpeed
    
    @staticmethod
    def getRpmFromSpeed(wheelSpeed: float, gearRatio: float, finalDrive: float, tireDiameter: float):
        wheelRps = wheelSpeed / (math.pi * tireDiameter)
        wheelRpm = wheelRps * 60
        engineRpm = wheelRpm * gearRatio * finalDrive
        return engineRpm
    
    @staticmethod
    def getEngineTorque0(engineRpm: float, throttle: float, maxTorque: float, maxTorqueRpm: float):
        return -(maxTorque / (maxTorqueRpm ** 2)) * throttle ** 2 * (1 / throttle * engineRpm - maxTorqueRpm) ** 2 + throttle ** 2 * maxTorque
    
    @staticmethod
    def getWheelTorque0(engineTorque: float, gearRatio: float, finalDrive: float):
        return engineTorque * gearRatio * finalDrive
    
    def __init__(self, options):
        self.options = options
        self.internal = Options({
            "rpm": 500,
            "speed": 2.23,
            "shiftProgress": 1,
            "gear": 0,
            "previousGear": 0,
            "throttle": 0.1,
            "brake": 0
        })
        self.metricsInternal = Options({
            "metrics": initHistory()
        })
    
    def __car_init__(self, car: Car):
        options = self.options
        internal = self.internal
        self.car = car
        options.transmission.__car_engine_init__(self)
    
    def getEngineTorque(self):
        options = self.options
        internal = self.internal
        engineTorque = CarEngine.getEngineTorque0(internal.rpm, internal.throttle, options.maxTorque, options.maxTorqueRpm)
        return engineTorque
    
    def getWheelTorque(self):
        options = self.options
        internal = self.internal
        gearRatio = self.getGearRatio()
        engineRpmNew = CarEngine.getRpmFromSpeed(internal.speed, gearRatio, options.finalDrive, options.tireDiameter)
        engineRpmOld = engineRpmNew * (options.gearRatios[internal.previousGear] / options.gearRatios[internal.gear])
        engineTorqueNew = CarEngine.getEngineTorque0(engineRpmNew, internal.throttle, options.maxTorque, options.maxTorqueRpm)
        engineTorqueOld = CarEngine.getEngineTorque0(engineRpmOld, internal.throttle, options.maxTorque, options.maxTorqueRpm)
        wheelTorqueNew = CarEngine.getWheelTorque0(engineTorqueNew, options.gearRatios[internal.gear], options.finalDrive)
        wheelTorqueOld = CarEngine.getWheelTorque0(engineTorqueOld, options.gearRatios[internal.previousGear], options.finalDrive)
        wheelTorque = map(internal.shiftProgress, 0, 1, wheelTorqueOld, wheelTorqueNew)
        return wheelTorque
    
    def getWheelForce(self):
        options = self.options
        internal = self.internal
        wheelTorque = self.getWheelTorque()
        wheelForce = wheelTorque / (options.tireDiameter / 2)
        return wheelForce
    
    def getWheelAcceleration(self):
        options = self.options
        internal = self.internal
        wheelForce = self.getWheelForce()
        wheelAcceleration = wheelForce / options.mass
        brakeForce = internal.brake * options.maximumBrakeForce
        brakeAcceleration = brakeForce / options.mass
        brakeDeceleration = (max(0, internal.speed - brakeAcceleration) - internal.speed)
        return wheelAcceleration + brakeDeceleration
    
    def getEnvironmentForce(self):
        options = self.options
        internal = self.internal
        car = self.car
        environment = car.environment
        windForce = -environment.getWindForce(options.frontalArea)
        return windForce
    
    def getEnvironmentAcceleration(self):
        options = self.options
        internal = self.internal
        environmentForce = self.getEnvironmentForce()
        environmentAcceleration = environmentForce / options.mass
        return environmentAcceleration
    
    def getAcceleration(self):
        options = self.options
        internal = self.internal
        wheelAcceleration = self.getWheelAcceleration()
        # environmentAcceleration = self.getEnvironmentAcceleration()
        environmentAcceleration = 0
        environmentAcceleration += (-1 * 0.5 * 1.204 * (internal.speed ** 2) * 0.27 * options.frontalArea) / options.mass
        environmentAcceleration += -1 * (0.01 * 1 * options.mass * 9.81) / options.mass
        return wheelAcceleration + environmentAcceleration
    
    def getGearRatio(self):
        options = self.options
        internal = self.internal
        gearRatio = CarEngine.getGearRatioTransition(internal.previousGear, internal.gear, internal.shiftProgress, options.gearRatios)
        return gearRatio
    
    def testShift(self, to):
        options = self.options
        internal = self.internal
        if to < 0 or to >= len(options.gearRatios):
            return False
        rpmCheck = CarEngine.getRpmFromSpeed(internal.speed, options.gearRatios[to], options.finalDrive, options.tireDiameter)
        if rpmCheck < options.rpmIdle:
            return False
        if rpmCheck > options.rpmLimit:
            return False
        return True
    
    def shiftTo(self, to):
        options = self.options
        internal = self.internal
        if not self.testShift(to):
            return
        if internal.shiftProgress == 1:
            internal.previousGear = internal.gear
        internal.shiftProgress = 0
        internal.gear = to
        self.updateRpm()
    
    def shiftUp(self):
        internal = self.internal
        self.shiftTo(internal.gear + 1)
    
    def shiftDown(self):
        internal = self.internal
        self.shiftTo(internal.gear - 1)
    
    def start(self):
        options = self.options
        internal = self.internal
        internal.gear = 1
        internal.rpm = options.rpmIdle
        self.updateSpeed()
    
    def updateRpm(self):
        options = self.options
        internal = self.internal
        gearRatio = self.getGearRatio()
        internal.rpm = CarEngine.getRpmFromSpeed(internal.speed, gearRatio, options.finalDrive, options.tireDiameter)
    
    def updateSpeed(self):
        options = self.options
        internal = self.internal
        gearRatio = self.getGearRatio()
        internal.speed = CarEngine.getSpeedFromRpm(internal.rpm, gearRatio, options.finalDrive, options.tireDiameter)
        
    def physicsLoop(self, timeDelta: float):
        options = self.options
        internal = self.internal
        metricsInternal = self.metricsInternal
        metrics = metricsInternal.metrics
        
        if internal.shiftProgress < 1:
            internal.shiftProgress += options.shiftTime * timeDelta
        if internal.shiftProgress > 1:
            internal.shiftProgress = 1
        if internal.rpm <= options.rpmIdle and internal.throttle <= options.throttleIdle:
           internal.throttle = options.throttleIdle
        
        acceleration = self.getAcceleration()
        internal.speed += acceleration * timeDelta
        self.updateRpm()
        
        if internal.rpm > options.rpmLimit:
            internal.rpm = options.rpmLimit
            self.updateSpeed()
        if internal.rpm < 0:
            internal.rpm = 0
            self.updateSpeed()
        
        metrics.pushHistory("Speed", internal.speed * 3.6)
        metrics.pushHistory("RPM", internal.rpm)
        metrics.pushHistory("Gear", internal.gear + 1)
        metrics.pushHistory("Previous Gear", internal.previousGear + 1)
        metrics.pushHistory("Gear Shift Progress", internal.shiftProgress * 100)
        metrics.pushHistory("Effective Gear Ratio", self.getGearRatio())
        metrics.pushHistory("Acceleration", self.getAcceleration())
        metrics.pushHistory("Engine Torque", self.getEngineTorque())
        metrics.pushHistory("Wheel Torque", self.getWheelTorque())
        metrics.pushHistory("Throttle", internal.throttle * 100)
        metrics.pushHistory("Brake", internal.brake * 100)
        
        options.transmission.physicsLoop(timeDelta)
    
    def drawLoop(self, timeDelta: float):
        options = self.options
        internal = self.internal
        
        options.transmission.drawLoop(timeDelta)
    
    def metricsLoop(self, timeDelta: float):
        options = self.options
        internal = self.internal
        metricsInternal = self.metricsInternal
        metrics = metricsInternal.metrics
        
        if imgui.begin_child("Engine"):
            metrics.drawGraphTable()
            
            options.transmission.metricsLoop(timeDelta)
        imgui.end_child()
    
    def controlLoop(self, timeDelta: float):
        options = self.options
        internal = self.internal
        
        if imgui.begin_child("Engine"):
            _, throttleValue = imgui.slider_float("Throttle", internal.throttle, v_min=0.01, v_max=1, format="%.2f")
            internal.throttle = throttleValue
            
            _, brakeValue = imgui.slider_float("Brake", internal.brake, v_min=0, v_max=1, format="%.2f")
            internal.brake = brakeValue
            
            stalled = internal.speed == 0
            if stalled:
                imgui.push_style_color(imgui.Col_.button, ImVec4(0.7, 0, 0, 1))
                imgui.push_style_color(imgui.Col_.button_hovered, ImVec4(0.7, 0.2, 0.2, 1))
            if imgui.button("Start"):
                self.start()
            if stalled:
                imgui.pop_style_color(2)
            
            options.transmission.controlLoop(timeDelta)
        imgui.end_child()
