from utils import *
from gui import *
from car import *

def main():
    currentDirectory = os.path.dirname(os.path.realpath(__file__))
    assetsDirectory = os.path.join(currentDirectory, "..", "assets")
    himgui.set_assets_folder(assetsDirectory)
    
    environment = Environment(Options({
        "scene": Scene(Options({
            "files": sorted_nicely([f for f in os.listdir(assetsDirectory) if f.startswith("bg_") and f.endswith(".png")]),
            "fitImage": 6,
            "aspect": 16 / 9,
            "segmentTotal": 300,
            "segmentGap": 0.574, # meters
            "segmentGapUv": 0.015625, # 29 pixels offset
            "segmentLength": 18.271 # meters
        }))
    }))
    mainCar = Car(Options({
        "files": sorted_nicely([f for f in os.listdir(assetsDirectory) if f.startswith("car_") and f.endswith(".png")]),
        "fitImage": 6,
        "aspect": 16 / 9,
        "segmentTotal": 300,
        "segmentGap": 0.1, # meters
        "segmentLength": 12, # meters
        "color": [random.random(), random.random(), random.random(), 1],
        "engine": CarEngine(Options({
            "gearRatios": [4.17, 2.34, 1.52, 1.14, 0.87, 0.69],
            "shiftTime": 0.5,
            "finalDrive": 3.9,
            "tireDiameter": 0.6319,
            "mass": 1460,
            "maxTorque": 200,
            "maxTorqueRpm": 3600,
            "rpmMaxPower": 4800,
            "rpmIdle": 900,
            "rpmLimit": 6500,
            "throttleIdle": 0.12,
            "frontalArea": 2.17,
            "maximumBrakeForce": 14600,
            "transmission": CarEngineAutoTransmissionR3(Options({}))
        })),
        "cruiseControl": CarCruiseControl(Options({
            "responseTime": 2 #seconds
        }))
    }))
    mainCar.__environment_init__(environment)
    simulation = Simulation(Options({
        "environment": environment,
        "mainCar": mainCar
    }))
    
    immapp.run(
        runner_params=himgui.RunnerParams(
            callbacks=himgui.RunnerCallbacks(
                show_gui=simulation.loop
            ),
            fps_idling=himgui.FpsIdling(
                enable_idling=False,
                is_idling=False,
                remember_enable_idling=False,
                fps_idle=60
            ),
            app_window_params=himgui.AppWindowParams(
                window_title="Advanced Cruise Control Simulation",
                restore_previous_geometry=True
            ),
            imgui_window_params=himgui.ImGuiWindowParams(
                remember_theme=True
            )
        ),
        add_ons_params=immapp.AddOnsParams(
            with_implot=True
        )
    )

class Simulation:
    def __init__(self, options):
        self.options = options
        self.internal = Options({
            "cars": [options.mainCar],
            "newCarOptions": Options({
                "summon": False,
                "distance": 200,
                "speed": 10,
                "gear": 0
            }),
            "lastTime": time.time(),
            "fpsTime": time.time(),
            "fpsCount": 0,
            "fps": 0
        })
        self.metricsInternal = Options({
            "metrics": initHistory()
        })
        options.environment.__init_simulation__(self)
    
    def loop(self):
        options = self.options
        internal = self.internal
        metricsInternal = self.metricsInternal
        metrics = metricsInternal.metrics
        
        now = time.time()
        timeDelta = now - internal.lastTime
        internal.lastTime = now
        internal.fpsCount += 1
        if internal.fpsCount >= 10:
            currentFps = internal.fpsCount / (now - internal.fpsTime)
            internal.fps = internal.fps * 0.8 + currentFps * 0.2
            internal.fpsTime = now
            internal.fpsCount = 0
        metrics.pushHistory("Time Delta", timeDelta * 1000)
        metrics.pushHistory("Immediate FPS", 1 / timeDelta)
        metrics.pushHistory("FPS", internal.fps)
        
        environment = options.environment
        scene = environment.options.scene
        cars = internal.cars
        environment.physicsLoop(timeDelta)
        for car in cars:
            car.physicsLoop(timeDelta)
        scene.internal.position = cars[0].internal.position
        
        imgui.dock_space(imgui.get_id("Dockspace"))
        
        imgui.push_style_var(StyleVar_.window_padding, (0, 0))
        imgui.begin("Visualisation")
        environment.drawLoop(timeDelta)
        for car in cars:
            car.drawLoop(timeDelta)
        imgui.end()
        imgui.pop_style_var()
        
        imgui.begin("Simulation Metrics")
        metrics.drawGraphTable()
        imgui.end()
        imgui.begin("Environment Metrics")
        environment.metricsLoop(timeDelta)
        imgui.end()
        imgui.begin("Main Car Metrics")
        cars[0].metricsLoop(timeDelta)
        imgui.end()
        
        for j in range(1, len(cars)):
            car = cars[j]
            if car.destroyed:
                continue
            imgui.begin(f"Car {j} Metrics")
            car.metricsLoop(timeDelta)
            imgui.end()
        
        imgui.begin("Environment Control")
        environment.controlLoop(timeDelta)
        imgui.end()
        imgui.begin("New Car Control")
        self.newCarControl()
        imgui.end()
        imgui.begin("Main Car Control")
        cars[0].controlLoop(timeDelta)
        imgui.end()
        
        for j in range(1, len(cars)):
            car = cars[j]
            if car.destroyed:
                continue
            imgui.begin(f"Car {j} Control")
            if imgui.button("Destroy"):
                self.destroyCar(j)
            car.controlLoop(timeDelta)
            imgui.end()
        self.newCarJob()
    
    def newCarControl(self):
        options = self.options
        internal = self.internal
        newCarOptions = internal.newCarOptions
        
        _, distanceValue = imgui.slider_float("Distance", newCarOptions.distance, v_min=10, v_max=1000, format="%.2f")
        newCarOptions.distance = distanceValue
        
        _, speedValue = imgui.slider_float("Speed", newCarOptions.speed, v_min=1, v_max=6, format="%.2f")
        newCarOptions.speed = speedValue
        
        _, gearValue = imgui.slider_int("Gear", newCarOptions.gear + 1, v_min=1, v_max=6)
        newCarOptions.gear = gearValue - 1
        
        if imgui.button("Summon"):
            newCarOptions.summon = True
    
    def newCarJob(self):
        options = self.options
        internal = self.internal
        environment = options.environment
        cars = internal.cars
        newCarOptions = internal.newCarOptions
        if not newCarOptions.summon:
            return
        
        newCar = self.createCar()
        newCar.__environment_init__(environment)
        cars.append(newCar)
        newCar.internal.position = cars[0].internal.position + newCarOptions.distance
        newCar.options.engine.internal.speed = newCarOptions.speed
        newCar.options.engine.internal.gear = newCarOptions.gear
        newCar.options.engine.updateRpm()
        newCar.destroyed = False
        newCarOptions.summon = False
        
    def destroyCar(self, i):
        options = self.options
        internal = self.internal
        environment = options.environment
        scene = environment.options.scene
        cars = internal.cars
        car = cars[i]
        car.destroyed = True
        car.options.engine.internal.throttle = 1
        oldPhysicsLoop = car.physicsLoop
        def overridePhysicsLoop(timeDelta):
            nonlocal oldPhysicsLoop, car, scene, cars
            oldPhysicsLoop(timeDelta)
            if car.internal.position - scene.internal.position < 500:
                return
            cars.remove(car)
        car.physicsLoop = overridePhysicsLoop
    
    def createCar(self):
        currentDirectory = os.path.dirname(os.path.realpath(__file__))
        assetsDirectory = os.path.join(currentDirectory, "..", "assets")
        return Car(Options({
            "files": sorted_nicely([f for f in os.listdir(assetsDirectory) if f.startswith("car_") and f.endswith(".png")]),
            "fitImage": 6,
            "aspect": 16 / 9,
            "segmentTotal": 300,
            "segmentGap": 0.1, # meters
            "segmentLength": 18.271, # meters
            "color": [random.random(), random.random(), random.random(), 1],
            "engine": CarEngine(Options({
                "gearRatios": [4.17, 2.34, 1.52, 1.14, 0.87, 0.69],
                "shiftTime": 0.5,
                "finalDrive": 3.9,
                "tireDiameter": 0.6319,
                "mass": 1460,
                "maxTorque": 200,
                "maxTorqueRpm": 3600,
                "rpmMaxPower": 4800,
                "rpmIdle": 900,
                "rpmLimit": 6500,
                "throttleIdle": 0.12,
                "frontalArea": 2.17,
                "maximumBrakeForce": 14600,
                "transmission": CarEngineAutoTransmissionAcceleration(Options({}))
            }))
        }))

if __name__ == "__main__":
    main()
